// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#ifndef _H_LOCALSIZEFIELD
#define _H_LOCALSIZEFIELD

#include "SizeFieldBase.h"
#include "DistanceFunction.h"

#include <set>
#include <utility>

// -------------------------------------------------------------------
/*
  This class stores a size field around an object. 
  The object is defined by a set of geometrical entities.
  The sizes are determined according to 3 values:
  - the size on the object
  - the thickness of the zone on which the local size field is acting
  - the size at the external boundary of that zone
  The value of the size at a point located in the zone is obtained by a 
  linear interpolation. Outside the zone, it returns a huge size.
*/
// -------------------------------------------------------------------

namespace MAd {

  // -------------------------------------------------------------------
  enum FUNCTION_TYPE {
    FT_LINEAR,
    FT_SQRT
  };

  class function {
  public:
    function() {}
    virtual FUNCTION_TYPE type() const = 0;
    virtual double y(double x) const = 0;
  };

  class linearFct : public function { // y = a*x + b
  public:
    linearFct(double _a, double _b): function(), a(_a), b(_b) {}
    virtual FUNCTION_TYPE type() const { return FT_LINEAR; };
    virtual double y(double x) const { return (a*x+b); }
  private:
    double a, b;
  };

  class sqrtFct : public function { // y = a*sqrt(x) + b
  public:
    sqrtFct(double _a, double _b): function(), a(_a), b(_b) {}
    virtual FUNCTION_TYPE type() const { return FT_SQRT; };
    virtual double y(double x) const { return (a*sqrt(x)+b); }
  private:
    double a, b;
  };

  // -------------------------------------------------------------------
  class LocalSizeField : public SizeFieldBase {

  public:

    LocalSizeField(pMesh m);
    LocalSizeField(const LocalSizeField& _lsf);
    ~LocalSizeField();

    sFieldType getType() const { return LOCALSFIELD; }

    void scale(double);

    void addGeometricEntity(int type, int tag);
    void updateTree();

    void setIsoSize(double rad, FUNCTION_TYPE type, 
                    double sizeIn, double sizeOut);
    void setAnisoSize(double rad, FUNCTION_TYPE type, 
                      double sizeIn, double sizeOut, double _tgSize);

    // get the size at a location (allocate space!)
    pMSize getSize(const pVertex) const;
    pMSize getSizeOnEntity(const pEntity, const double[3]) const;

    // edge length (squared)
    double SF_VV_lengthSq(const pVertex, const pVertex) const;
    double SF_XYZ_lengthSq(const double[3], const double[3],
                           const pMSize, const pMSize=NULL) const;

    // face area (squared)
    double SF_F_areaSq(const pFace) const;
    double SF_XYZ_areaSq(const double[3][3], const pMSize,
                         const double[3]) const;

    // region volume
    double SF_R_volume(const pRegion) const;
    double SF_XYZ_volume(const double[4][3], const pMSize) const;

    // center and its associated size
    double SF_E_center(const pEdge, double[3], double * reducSq, pMSize *) const;
    double SF_VV_center(const pVertex, const pVertex,
                        double[3], double * reducSq, pMSize *) const;

  private:

    std::set<pVertex> collectVertices() const;
    void addVerticesOnGEntity(int type, int tag, std::set<pVertex> * verts) const;

    double getDistance(const double[3]) const;

  private:

    pMesh mesh;

    std::set<std::pair<int,int> > geomEntities; // set of (type,tag) 

    bool isotropic;
    double radius;
    function * sFct;
    double tgSize;

    distanceFunction dFct;
  };

}

// -------------------------------------------------------------------
#endif
