// -*- C++ -*-
// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "DistanceFunction.h"
#include "MAdOutput.h"
#include "CallBackManager.h"

namespace MAd {

  // -------------------------------------------------------------------
  void DistFct_CBMoveFct(pVertex pv, double * target, void * data)
  {
    distanceFunction * dFct = (distanceFunction *) data;
    dFct->clearVertexData(pv);
  }

  // -------------------------------------------------------------------
  void DistFct_CBFct(pPList before, pPList after, void * data,
                     operationType type , pEntity ppp)
  {
    distanceFunction * dFct = (distanceFunction *) data;
    
    switch (type) {
    case MAd_ESPLIT: {
      // In the edge split case, we have to interpolate the data at the new node

      if ( dFct->gradientAvailable() )
        {
          double grads[2][3];
          
          // find the old edge
          void * tmp = 0;
          pEdge pe = (pEdge)PList_next(before,&tmp);
          double t = E_linearParams(pe,(pVertex)ppp);
          
          // get datas at nodes
          dFct->getGradient( E_vertex((pEdge)pe, 0), grads[0] );
          dFct->getGradient( E_vertex((pEdge)pe, 1), grads[1] );

          double * newGrad = new double[3];
          for (int i=0; i<3; i++) newGrad[i] = (1.-t) * grads[0][i] + t * grads[1][i];

          // attach it
          dFct->attachGradient((pVertex)ppp, newGrad);
        }
      break;
    } 
    case MAd_ECOLLAPSE: {
      // In the edge collapse case, we have to delete the datas attached to the deleted node
      dFct->clearVertexData((pVertex)ppp);
      break;
    }
    case MAd_ESWAP:
    case MAd_FSWAP: {
      // nothing to be done (no modification at nodes)
      break;
    }
    case MAd_RREMOVE: {
      void * temp = NULL;
      while ( pEntity pE = PList_next(before,&temp) ) {
        if ( EN_type(pE) == 0 ) {
          dFct->clearVertexData((pVertex)pE);
        }
      }
      break;
    }
    case MAd_UNKNOWNOPERATION:
    case MAd_VERTEXMOVE:
    case MAd_FCOLLAPSE:
    case MAd_DESPLTCLPS:
    default: {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Not implemented for mesh modification %d", type);
    }
    }
  }

  // -------------------------------------------------------------------
  distanceFunction::distanceFunction(pMesh m): mesh(m), gradComputed(false)
  {
    distId = MD_newMeshDataId();
    vGradId = MD_newMeshDataId();
    rGradId = MD_newMeshDataId();
    kdSearch = new SearchTool();

    CallBackManagerSgl::instance().registerCallBackMove(DistFct_CBMoveFct,this);
    CallBackManagerSgl::instance().registerCallBack(DistFct_CBFct,this);
  }

  // -------------------------------------------------------------------
  distanceFunction::~distanceFunction()
  {
    clearDistances();
    MD_deleteMeshDataId(distId);

    clearGradientAtVertices();
    MD_deleteMeshDataId(vGradId);

    clearGradientInElements();
    MD_deleteMeshDataId(rGradId);

    if (kdSearch) delete kdSearch;

    CallBackManagerSgl::instance().unregisterCallBackMove(DistFct_CBMoveFct,this);
    CallBackManagerSgl::instance().unregisterCallBack(DistFct_CBFct,this);
  }
  
  // -------------------------------------------------------------------
  void distanceFunction::computeTree(const std::set<pVertex>& vertices)
  {
    kdSearch->reset();

    int nbVert = vertices.size();
    kdSearch->allocatePoints(nbVert);

    std::set<pVertex> :: const_iterator itV    = vertices.begin();
    std::set<pVertex> :: const_iterator itVEnd = vertices.end();
    int k = 0;
    for ( ; itV != itVEnd ; ++itV)
      {
        double coord[3]; V_coord(*itV,coord);
        kdSearch->addPoint(k,coord);
        k++;
      }

    kdSearch->allocateTree(nbVert);

    clearDistances();
    clearGradientAtVertices();
    clearGradientInElements();
    
    gradComputed = false;
  }
  
  // -------------------------------------------------------------------
  void distanceFunction::computeAllDistances() const
  {
    pVertex pv;
    VIter vit = M_vertexIter(mesh);
    while ( ( pv = VIter_next(vit) ) )
      {
        computeDistance(pv);
      }
    VIter_delete(vit);
  }

  // -------------------------------------------------------------------
  double distanceFunction::computeDistance(pVertex pv) const
  {
    double xyz[3]; V_coord(pv,xyz);
    double dist = sqrt(kdSearch->computeDistanceSq(xyz));
    EN_attachDataDbl((pEntity)pv,distId,dist);
    return dist;
  }

  // -------------------------------------------------------------------
  double distanceFunction::computeDistance(const double xyz[3]) const
  {
    return sqrt(kdSearch->computeDistanceSq(xyz));
  }

  // -------------------------------------------------------------------
  double distanceFunction::computeDistSq(const double xyz[3]) const
  {
    return kdSearch->computeDistanceSq(xyz);
  }

  // -------------------------------------------------------------------
  double distanceFunction::getDistance(pVertex pv) const
  {
    double dist;
    if ( EN_getDataDbl((pEntity)pv,distId,&dist) ) return dist;
    return computeDistance(pv);
  }

  // -------------------------------------------------------------------
  void distanceFunction::clearDistances() const
  {
    pVertex pv;
    VIter vit = M_vertexIter(mesh);
    while ( ( pv = VIter_next(vit) ) )
      {
        EN_deleteData((pEntity)pv,distId);
      }
    VIter_delete(vit);
  }

  // -------------------------------------------------------------------
  void distanceFunction::computeGradientInElements() const
  {
    double gsf[4][3] = {
      {-1., -1., -1.}, 
      { 1.,  0.,  0.}, 
      { 0.,  1.,  0.}, 
      { 0.,  0.,  1.}, 
    };

    pRegion pr;
    RIter rit = M_regionIter(mesh);
    while ( ( pr = RIter_next(rit) ) )
      {
        pPList rVerts = R_vertices(pr);
        double dist[4];
        void * temp = NULL;
        pVertex pv;
        int i = 0;
        while ( ( pv = (pVertex)PList_next(rVerts,&temp) ) )
          {
            dist[i] = getDistance(pv);
            i++;
          }
        PList_delete(rVerts);

        double ijac[3][3], detj;
        detj = R_invJacobian(pr,ijac);

        void * del;
        if ( EN_getDataPtr((pEntity)pr,rGradId,&del) && del ) delete [] (double*)del;
        double * grad = new double[3];
        for ( int iC=0; iC<3; iC++)
          {
            grad[iC] = 0.;
            for (int iSF=0; iSF<4; iSF++) {
              grad[iC] += dist[iSF] * ( gsf[iSF][0] * ijac[0][iC] +
                                        gsf[iSF][1] * ijac[1][iC] +
                                        gsf[iSF][2] * ijac[2][iC]   );
            }
          }

        EN_attachDataPtr((pEntity)pr,rGradId,grad);
      }
    RIter_delete(rit);
  }

  // -------------------------------------------------------------------
  void distanceFunction::clearGradientInElements() const
  {
    pRegion pr;
    RIter rit = M_regionIter(mesh);
    while ( ( pr = RIter_next(rit) ) )
      {
        EN_deleteData((pEntity)pr,rGradId);
      }
    RIter_delete(rit);
  }

  // -------------------------------------------------------------------
  void distanceFunction::computeGradientAtVertices()
  {
    computeGradientInElements();

    pVertex pv;
    VIter vit = M_vertexIter(mesh);
    while ( ( pv = VIter_next(vit) ) )
      {
        void * del;
        if ( EN_getDataPtr((pEntity)pv,vGradId,&del) && del ) delete [] (double*)del;
        double * grad = new double[3];
        for (int i=0; i<3; i++) grad[i] = 0.;

        pPList vRegs = V_regions(pv);
        void * temp = NULL;
        void * rGrad = NULL;
        pRegion pr;
        while ( pr = (pRegion)PList_next(vRegs,&temp) )
          {
            if ( !EN_getDataPtr((pEntity)pr,rGradId,&rGrad) ) throw;
            grad[0] += ((double*)rGrad)[0];
            grad[1] += ((double*)rGrad)[1];
            grad[2] += ((double*)rGrad)[2];
          }

        int nRegs = PList_size(vRegs);
        PList_delete(vRegs);
        double invNRegs = 1. / (double)nRegs;
        for (int i=0; i<3; i++) grad[i] *= invNRegs;
        
        EN_attachDataPtr((pEntity)pv,vGradId,grad);
      }
    VIter_delete(vit);

    clearGradientInElements();

    gradComputed = true;
  }

  // -------------------------------------------------------------------
  void distanceFunction::clearGradientAtVertices()
  {
    pVertex pv;
    VIter vit = M_vertexIter(mesh);
    while ( ( pv = VIter_next(vit) ) )
      {
        EN_deleteData((pEntity)pv,vGradId);
      }
    VIter_delete(vit);

    gradComputed = false;
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradient(const pVertex pv, double grad[3]) const
  {
    void * tmp;
    if ( !EN_getDataPtr((pEntity)pv,vGradId,&tmp) ) return false;
    for (int i=0; i<3; i++) grad[i] = ((double*) tmp)[i];
    return true;
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnEntity(const pEntity entity, 
                                             const double xyz[3],
                                             double grad[3]) const
  {
    int type = EN_type(entity);
    switch(type) {
    case 0: return getGradient((pVertex)entity,grad);
    case 1: return getGradientOnEdge((pEdge)entity,xyz,grad);
    case 2: return getGradientOnFace((pFace)entity,xyz,grad);
    case 3: return getGradientOnRegion((pRegion)entity,xyz,grad);
    default: throw;
    }
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnEdge(const pEdge edge, 
                                           const double xyz[3],
                                           double grad[3]) const
  {
    double u = E_linearParams(edge,xyz);
    return getGradientOnEdgeParam(edge,u,grad);
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnEdgeParam(const pEdge pE, 
                                                const double u,
                                                double grad[3]) const
  {
    double g0[3], g1[3];
    if ( !getGradient(E_vertex(pE,0), g0) ||
         !getGradient(E_vertex(pE,1), g1) ) return false;
  
    for (int i=0; i<3; i++) grad[i] = (1.-u) * g0[i] + u * g1[i];
    return true;
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnFace(const pFace face, 
                                           const double xyz[3],
                                           double grad[3]) const
  {
    double u[2];
    F_linearParams(face,xyz,u);
    return getGradientOnFaceParam(face,u,grad);
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnFaceParam(const pFace face, 
                                                const double u[2],
                                                double grad[3]) const
  {
    double g0[3], g1[3], g2[3];
    if ( !getGradient(F_vertex(face,0), g0) ||
         !getGradient(F_vertex(face,1), g1) ||
         !getGradient(F_vertex(face,2), g2) ) return false;
  
    for (int i=0; i<3; i++) grad[i] = (1.-u[0]-u[1]) * g0[i] + u[0] * g1[i] + u[1] * g2[i];
    return true;
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnRegion(const pRegion region, 
                                             const double xyz[3],
                                             double grad[3]) const
  {
    double u[3];
    R_linearParams(region, xyz, u);
    return getGradientOnRegionParam(region, u, grad);
  }

  // -------------------------------------------------------------------
  bool distanceFunction::getGradientOnRegionParam(const pRegion region, 
                                                  const double u[3],
                                                  double grad[3]) const
  {
    double g0[3], g1[3], g2[3], g3[3];
    if ( !getGradient(R_vertex(region,0), g0) ||
         !getGradient(R_vertex(region,1), g1) ||
         !getGradient(R_vertex(region,2), g2) ||
         !getGradient(R_vertex(region,3), g3) ) return false;
  
    for (int i=0; i<3; i++) {
      grad[i] = (1.-u[0]-u[1]-u[2]) * g0[i] + u[0] * g1[i] + u[1] * g2[i] + u[2] * g3[i];
    }
    return true;
  }

  // -------------------------------------------------------------------
  void distanceFunction::attachGradient(pVertex pv, double grad[3]) const
  {
    EN_attachDataPtr((pEntity)pv,vGradId,grad);
  }

  // -------------------------------------------------------------------
  void distanceFunction::outputDistance(const char * fn) const
  {
    MAdAttachedNodalDataOutput(mesh, fn, distId);
  }

  // -------------------------------------------------------------------
  void distanceFunction::outputGradAtVertices(const char * fn) const
  {
    if ( !gradComputed ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Distance gradients are not computed");
    }
    MAdAttachedNodalDataVecOutput(mesh, fn, vGradId);
  }

  // -------------------------------------------------------------------
  void distanceFunction::clearVertexData(pVertex pv) const
  {
    EN_deleteData((pEntity)pv, vGradId);
    EN_deleteData((pEntity)pv, distId);
  }

  // -------------------------------------------------------------------
}
